def foo() -> str:
    """Summary line.

    Extended description of function.

    Args:
        foo (str): Description of arg1

    Returns:
        str: Description of return value
    """

    return "foo"


if __name__ == "__main__":  # pragma: no cover
    pass
